import logging
import pprint

import acquire
import pytest
from acquire import DeviceKind, SampleType
from acquire.acquire import SignalIOKind, Trigger, TriggerEdge


@pytest.fixture(scope="module")
def runtime():
    runtime = acquire.Runtime()
    yield runtime


def test_vieworks_camera_is_preset(runtime: acquire.Runtime):
    dm = runtime.device_manager()
    assert dm.select(DeviceKind.Camera, "VIEWORKS.*")


def test_vieworks_stream(
    runtime: acquire.Runtime, request: pytest.FixtureRequest
):
    dm = runtime.device_manager()
    p = runtime.get_configuration()

    p.video[0].camera.identifier = dm.select(DeviceKind.Camera, "VIEWORKS.*")
    assert p.video[0].camera.identifier

    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Zarr")
    assert p.video[0].storage.identifier
    p.video[0].storage.settings.filename = request.node.name + ".zarr"
    p.video[0].storage.settings.bytes_per_chunk = 1 << 30  # 1 GB chunks

    # Set the camera here so we can query it's triggering capabilities.
    # This comes in the form of the returned properties.
    p = runtime.set_configuration(p)

    p.video[0].camera.settings.binning = 1
    p.video[0].camera.settings.shape = (14192, 10640)
    p.video[0].camera.settings.pixel_type = SampleType.U12
    p.video[0].camera.settings.exposure_time_us = 1e3

    p.video[0].storage.settings.pixel_scale_um = (0.2, 0.2)
    p.video[0].max_frame_count = 10

    p = runtime.set_configuration(p)

    logging.info(pprint.pformat(p.dict()))

    runtime.start()
    runtime.stop()


def test_vieworks_configure_triggering(runtime: acquire.Runtime):
    dm = runtime.device_manager()
    p = runtime.get_configuration()

    p.video[0].camera.identifier = dm.select(DeviceKind.Camera, "VIEWORKS.*")
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    assert p.video[0].camera.identifier

    p = runtime.set_configuration(p)

    # When the camera is first selected, triggers should be disabled
    assert not p.video[0].camera.settings.input_triggers.frame_start.enable

    #
    # Enable Line0:
    #
    # There's really own two things to set. On the VP-151MX, there's only
    # one kind of event that can be triggered - the frame exposure start.
    p.video[0].camera.settings.input_triggers.frame_start = Trigger(
        enable=True, line=0, edge="Rising"
    )
    assert p.video[0].camera.settings.input_triggers.frame_start.enable

    p = runtime.set_configuration(p)
    assert p.video[0].camera.settings.input_triggers.frame_start.enable
    assert p.video[0].camera.settings.input_triggers.frame_start.line == 0

    #
    # Enable Software triggering:
    #
    # There's really own two things to set. On the VP-151MX, there's only
    # one kind of event that can be triggered - the frame exposure start.
    p.video[0].camera.settings.input_triggers.frame_start = Trigger(
        enable=True, line=1, edge="Rising"
    )

    p = runtime.set_configuration(p)

    assert p.video[0].camera.settings.input_triggers.frame_start.enable
    assert p.video[0].camera.settings.input_triggers.frame_start.line == 1

    #
    # Disable triggering:
    #
    p.video[0].camera.settings.input_triggers.frame_start.enable = False
    p = runtime.set_configuration(p)



from typing import (
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    final,
    overload,
)

from numpy.typing import NDArray

@final
class AvailableData:
    def frames(self) -> Iterator[VideoFrame]: ...
    def get_frame_count(self) -> int: ...
    def __iter__(self) -> Iterator[VideoFrame]: ...

@final
class Camera:
    identifier: Optional[DeviceIdentifier]
    settings: CameraProperties
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...

@final
class CameraProperties:
    exposure_time_us: float
    line_interval_us: float
    binning: float
    pixel_type: SampleType
    readout_direction: Direction
    offset: Tuple[int, int]
    shape: Tuple[int, int]
    input_triggers: InputTriggers
    output_triggers: OutputTriggers
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...

@final
class DeviceIdentifier:
    id: Tuple[int, int]
    kind: DeviceKind
    name: str
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...
    @staticmethod
    def none() -> DeviceIdentifier: ...
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class DeviceKind:
    Camera: ClassVar[DeviceKind] = DeviceKind.Camera
    NONE: ClassVar[DeviceKind] = DeviceKind.NONE
    Signals: ClassVar[DeviceKind] = DeviceKind.Signals
    StageAxis: ClassVar[DeviceKind] = DeviceKind.StageAxis
    Storage: ClassVar[DeviceKind] = DeviceKind.Storage
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class DeviceManager:
    def devices(self) -> List[DeviceIdentifier]: ...
    @overload
    def select(self, kind: DeviceKind) -> Optional[DeviceIdentifier]: ...
    @overload
    def select(
        self, kind: DeviceKind, name: Optional[str]
    ) -> Optional[DeviceIdentifier]: ...
    def select_one_of(
        self, kind: DeviceKind, names: List[str]
    ) -> Optional[DeviceIdentifier]: ...

@final
class DeviceState:
    Closed: ClassVar[DeviceState] = DeviceState.Closed
    AwaitingConfiguration: ClassVar[
        DeviceState
    ] = DeviceState.AwaitingConfiguration
    Armed: ClassVar[DeviceState] = DeviceState.Armed
    Running: ClassVar[DeviceState] = DeviceState.Running
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class Direction:
    Backward: ClassVar[Direction] = Direction.Backward
    Forward: ClassVar[Direction] = Direction.Forward
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class InputTriggers:
    acquisition_start: Trigger
    exposure: Trigger
    frame_start: Trigger
    def dict(self) -> Dict[str, Any]: ...

@final
class OutputTriggers:
    exposure: Trigger
    frame_start: Trigger
    trigger_wait: Trigger
    def dict(self) -> Dict[str, Any]: ...

@final
class PID:
    derivative: float
    integral: float
    proportional: float
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...

@final
class Properties:
    video: Tuple[VideoStream, VideoStream]
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...

@final
class Runtime:
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def device_manager(self) -> DeviceManager: ...
    def get_available_data(self, stream_id: int) -> AvailableData: ...
    def get_configuration(self) -> Properties: ...
    def get_state(self) -> DeviceState: ...
    def set_configuration(self, properties: Properties) -> Properties: ...
    def start(self) -> None: ...
    def stop(self) -> None: ...
    def abort(self) -> None: ...

@final
class SampleRateHz:
    numerator: int
    denominator: int
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...

@final
class SampleType:
    F32: ClassVar[SampleType] = SampleType.F32
    I16: ClassVar[SampleType] = SampleType.I16
    I8: ClassVar[SampleType] = SampleType.I8
    U16: ClassVar[SampleType] = SampleType.U16
    U8: ClassVar[SampleType] = SampleType.U8
    U10: ClassVar[SampleType] = SampleType.U10
    U12: ClassVar[SampleType] = SampleType.U12
    U14: ClassVar[SampleType] = SampleType.U14
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


@final
class SignalIOKind:
    Input: ClassVar[SignalIOKind] = SignalIOKind.Input
    Output: ClassVar[SignalIOKind] = SignalIOKind.Output
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class SignalType:
    Analog: ClassVar[SignalType] = SignalType.Analog
    Digital: ClassVar[SignalType] = SignalType.Digital
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class Storage:
    identifier: Optional[DeviceIdentifier]
    settings: StorageProperties
    def dict(self) -> Dict[str, Any]: ...

@final
class StorageProperties:
    bytes_per_chunk: int
    external_metadata_json: Optional[str]
    filename: Optional[str]
    first_frame_id: int
    pixel_scale_um: Tuple[float, float]
    def dict(self) -> Dict[str, Any]: ...

@final
class Trigger:
    edge: TriggerEdge
    enable: bool
    line: int
    kind: SignalIOKind
    def __init__(self, *args: None, **kwargs: Any) -> None: ...
    def dict(self) -> Dict[str, Any]: ...

@final
class TriggerEdge:
    Falling: ClassVar[TriggerEdge] = TriggerEdge.Falling
    NotApplicable: ClassVar[TriggerEdge] = TriggerEdge.NotApplicable
    Rising: ClassVar[TriggerEdge] = TriggerEdge.Rising
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __int__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

@final
class VideoFrame:
    def data(self) -> NDArray[Any]: ...
    def metadata(self) -> VideoFrameMetadata: ...

@final
class VideoFrameMetadata:
    frame_id: int
    timestamps: VideoFrameTimestamps
    def dict(self) -> Dict[str, Any]: ...

@final
class VideoFrameTimestamps:
    hardware: int
    acq_thread: int
    def dict(self) -> Dict[str, Any]: ...

@final
class VideoStream:
    camera: Camera
    storage: Storage
    max_frame_count: int
    frame_average_count: int
    def dict(self) -> Dict[str, Any]: ...

@final
class VoltageRange:
    mn: float
    mx: float
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, mn: float, mx: float) -> None: ...
    def dict(self) -> Dict[str, float]: ...

def core_api_version() -> str: ...



import time
from typing import Any, List, Optional, Tuple, Union

from . import acquire
from .acquire import *

__doc__ = acquire.__doc__

import logging

FORMAT = (
    "%(levelname)s %(name)s %(asctime)-15s %(filename)s:%(lineno)d %(message)s"
)
logging.basicConfig(format=FORMAT)
logging.getLogger().setLevel(logging.INFO)


def setup(
    runtime: Runtime,
    camera: Union[str, List[str]] = "simulated: radial sin",
    storage: Union[str, List[str]] = "Tiff",
    output_filename: Optional[str] = "out.tif",
) -> Properties:
    def normalize_fallback_arg(arg: Union[str, List[str]]) -> List[str]:
        if isinstance(arg, str):
            return [arg]
        return arg

    camera = normalize_fallback_arg(camera)
    storage = normalize_fallback_arg(storage)

    dm = runtime.device_manager()
    p = runtime.get_configuration()

    p.video[0].camera.identifier = dm.select_one_of(DeviceKind.Camera, camera)
    p.video[0].camera.settings.binning = 1
    p.video[0].camera.settings.shape = (640, 480)
    p.video[0].storage.identifier = dm.select_one_of(
        DeviceKind.Storage, storage
    )
    p.video[0].storage.settings.filename = output_filename
    p.video[0].max_frame_count = 100
    p.video[0].frame_average_count = 0  # disables

    return p


def setup_one_streams(runtime: Runtime, frame_count: int) -> Properties:
    dm = runtime.device_manager()
    p = runtime.get_configuration()

    cameras = [
        d.name
        for d in dm.devices()
        if (d.kind == DeviceKind.Camera) and ("C15440" in d.name)
    ]
    logging.warning(f"Cameras {cameras}")

    p.video[0].camera.identifier = dm.select(DeviceKind.Camera, cameras[0])
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    # p.video[0].storage.settings.filename = output_filename
    p.video[0].camera.settings.binning = 1
    p.video[0].camera.settings.shape = (2304, 2304)
    p.video[0].camera.settings.pixel_type = SampleType.U16
    p.video[0].max_frame_count = frame_count
    p.video[0].frame_average_count = 0  # disables

    return p


def setup_two_streams(runtime: Runtime, frame_count: int) -> Properties:
    dm = runtime.device_manager()
    p = runtime.get_configuration()

    cameras = [
        d.name
        for d in dm.devices()
        if (d.kind == DeviceKind.Camera) and ("C15440" in d.name)
    ]
    if len(cameras) < 2:
        cameras = ["simulated.*random.*", "simulated.*sin.*"]
    logging.warning(f"Cameras {cameras}")

    p.video[0].camera.identifier = dm.select(DeviceKind.Camera, cameras[0])
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    # p.video[0].storage.settings.filename = output_filename
    p.video[0].camera.settings.binning = 1
    p.video[0].camera.settings.shape = (2304, 2304)
    p.video[0].camera.settings.pixel_type = SampleType.U16
    p.video[0].max_frame_count = frame_count
    p.video[0].frame_average_count = 0  # disables

    p.video[1].camera.identifier = dm.select(DeviceKind.Camera, cameras[1])
    p.video[1].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    # p.video[1].storage.settings.filename = output_filename
    p.video[1].camera.settings.binning = 1
    p.video[1].camera.settings.shape = (64, 64)
    p.video[1].camera.settings.pixel_type = SampleType.U16
    p.video[1].max_frame_count = frame_count
    p.video[1].frame_average_count = 0  # disables

    return p


g_runtime = None


def gui(
    viewer: "napari.Viewer",
    frame_count: int = 100,
):
    """Napari dock-widget plugin entry-point

    This instances a magicgui dock widget that streams video to a layer.
    """
    import numpy.typing as npt
    from napari.qt.threading import thread_worker
    from numpy import cumsum, histogram, where

    update_times: List[float] = []

    def get_runtime():
        global g_runtime
        if g_runtime is None:
            logging.info("INITING RUNTIME")
            g_runtime = acquire.Runtime()
        else:
            logging.info("REUSING RUNTIME")
        return g_runtime

    def update_layer(args: Tuple[npt.NDArray[Any], int]):
        (new_image, stream_id) = args
        layer_key = f"Video {stream_id}"
        try:
            clock = time.time()

            layer = viewer.layers[layer_key]
            layer._slice.image._view = new_image
            layer.events.set_data()
            # layer.data=new_image # public api adds another 1-2 ms/frame

            elapsed = time.time() - clock
            update_times.append(elapsed)
            logging.info(f"UPDATED LAYER {layer_key} in {elapsed} s")
        except KeyError:
            # (nclack) This takes ~ 60ms for 630x480 the one time I measured it
            viewer.add_image(new_image, name=layer_key)

    @thread_worker(connect={"yielded": update_layer})
    def do_acquisition():
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)

        runtime = get_runtime()
        # p = setup_two_streams(runtime,frame_count)
        p = setup_two_streams(runtime, frame_count)

        p = runtime.set_configuration(p)

        runtime.start()

        nframes = [0, 0]
        stream_id = 0

        def is_not_done() -> bool:
            return (nframes[0] < p.video[0].max_frame_count) or (
                nframes[1] < p.video[1].max_frame_count
            )

        while is_not_done():  # runtime.get_state()==DeviceState.Running:
            clock = time.time()

            if nframes[stream_id] < p.video[stream_id].max_frame_count:

                if packet := runtime.get_available_data(stream_id):
                    n = packet.get_frame_count()
                    f = next(packet.frames())
                    im = f.data().squeeze().copy()
                    logging.debug(
                        f"stream {stream_id} frame {f.metadata().frame_id}"
                    )

                    # TODO: (nclack) fix this awkwardness.
                    f = None
                    packet = None

                    yield (im, stream_id)

                    nframes[stream_id] += n
                    logging.info(
                        f"[stream {stream_id}] frame count: {nframes}"
                    )
            stream_id = (stream_id + 1) % 2

            elapsed = time.time() - clock
            time.sleep(max(0, 0.03 - elapsed))
        logging.info("stopping")

        counts, bins = histogram(update_times)
        p50 = bins[where(cumsum(counts) >= 0.5 * len(update_times))[0][0]]
        p90 = bins[where(cumsum(counts) >= 0.9 * len(update_times))[0][0]]
        logging.info(f"Update times - median: {p50*1e3} ms  90%<{p90*1e3} ms")

        runtime.stop()
        logging.info("STOPPED")

    do_acquisition()


# TODO: (nclack) add context manager around runtime and start/stop



import logging

import acquire
import pytest
from acquire import DeviceKind, SampleType


@pytest.fixture(scope="module")
def runtime():
    runtime = acquire.Runtime()
    yield runtime


def test_ext_triggering(runtime: acquire.Runtime):
    dm = runtime.device_manager()
    p = runtime.get_configuration()

    cameras = [
        d.name
        for d in dm.devices()
        if (d.kind == DeviceKind.Camera) and ("C15440" in d.name)
    ]
    assert len(cameras) > 0, "No C15440 cameras found"
    logging.warning(f"Cameras {cameras}")

    p.video[0].camera.identifier = dm.select(DeviceKind.Camera, cameras[0])
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    p.video[0].camera.settings.binning = 1
    p.video[0].camera.settings.shape = (2304, 2304)
    p.video[0].camera.settings.pixel_type = SampleType.U16
    p.video[0].max_frame_count = 1
    p.video[0].frame_average_count = 0  # disables

    # Set the camera here so we can query it's triggering capabilities.
    # This comes in the form of the returned properties.
    p.video[0].camera.settings.input_triggers.frame_start = acquire.Trigger(
        enable=True, line=0, edge="Rising"
    )
    # Call set_configuration() again to apply the trigger properties
    p = runtime.set_configuration(p)
    # Check that it was accepted
    assert p.video[0].camera.settings.input_triggers.frame_start.enable is True



import json
import logging
import math
import sys
import time
from time import sleep
from typing import Any, Dict, List

import acquire
import dask.array as da
import numcodecs.blosc as blosc
import pytest
import tifffile
import zarr
from acquire.acquire import DeviceKind, DeviceState, Runtime, Trigger
from ome_zarr.io import parse_url
from ome_zarr.reader import Reader


@pytest.fixture(scope="module")
def runtime():
    runtime = acquire.Runtime()
    yield runtime


def test_set():
    t = Trigger()
    assert not t.enable
    t.enable = True
    assert t.enable


def test_list_devices(runtime: Runtime):
    dm = runtime.device_manager()
    for d in dm.devices():
        print(d.dict())


def test_set_camera_identifier(runtime: Runtime):
    dm = runtime.device_manager()

    p = runtime.get_configuration()
    assert (
        p.video[0].camera.identifier is not None
        and p.video[0].camera.identifier.kind == acquire.DeviceKind.NONE
    )
    p.video[0].camera.identifier = dm.select(
        acquire.DeviceKind.Camera, "simulated: radial sin"
    )
    assert p.video[0].camera.identifier is not None


@pytest.mark.parametrize(
    "input,expected",
    [
        (["does not exist 1", "does not exist 2", "does not exist 3"], None),
        (
            [
                "does not exist 1",
                "simulated: radial sin",
                "simulated: uniform random",
            ],
            "simulated: radial sin",
        ),
        (["simulated: radial sin"], "simulated: radial sin"),
        (["simulated.*sin"], "simulated: radial sin"),
        ([".*radial.*"], "simulated: radial sin"),
        ([], None),
    ],
)
def test_select_one_of(
    runtime: Runtime,
    input: List[str],
    expected: str,
):
    h = runtime.device_manager().select_one_of(DeviceKind.Camera, input)
    result = None if h is None else h.name
    assert result == expected


def test_select_empty_string(runtime: Runtime):
    assert runtime.device_manager().select(DeviceKind.Storage, "")


def test_zero_conf_start(runtime: Runtime):
    with pytest.raises(RuntimeError):
        runtime.start()


def test_repeat_acq(runtime: Runtime):
    p = acquire.setup(runtime, "simulated: radial sin", "Trash")
    assert p.video[0].camera.identifier is not None
    assert p.video[0].storage.identifier is not None
    assert p.video[0].storage.settings.filename == "out.tif"
    p.video[0].camera.settings.shape = (192, 108)
    p.video[0].max_frame_count = 10
    p = runtime.set_configuration(p)
    runtime.start()
    while True:
        if a := runtime.get_available_data(0):
            logging.info(f"Got {a.get_frame_count()}")
            break
    runtime.stop()
    assert runtime.get_available_data(0) is None
    # TODO: (nclack) assert 1 acquired frame. stop should block
    runtime.start()
    while True:
        if a := runtime.get_available_data(0):
            logging.info(f"Got {a.get_frame_count()}")
            break
    runtime.stop()
    assert runtime.get_available_data(0) is None
    # TODO: (nclack) assert 1 more acquired frame. stop cancels and waits.


def test_repeat_with_no_stop(runtime: Runtime):
    """Stop is required between starts. This tests that an exception is
    raised."""
    p = acquire.setup(runtime, "simulated: radial sin", "Trash")
    assert p.video[0].camera.identifier is not None
    assert p.video[0].storage.identifier is not None
    p.video[0].camera.settings.shape = (192, 108)
    p.video[0].max_frame_count = 11
    p = runtime.set_configuration(p)
    runtime.start()
    # wait for 1 frame
    while True:
        if a := runtime.get_available_data(0):
            logging.info(f"Got {a.get_frame_count()}")
            a = None
            break
    # acq is still on going here
    with pytest.raises(RuntimeError):
        logging.info("Next start should fail gracefully")
        runtime.start()
    runtime.stop()


def test_set_storage(runtime: Runtime):
    dm = runtime.device_manager()

    p = runtime.get_configuration()
    p.video[0].storage.identifier = None
    p = runtime.set_configuration(p)
    assert p.video[0].storage.identifier is not None
    assert p.video[0].storage.identifier.kind == acquire.DeviceKind.NONE
    p.video[0].storage.identifier = dm.select(
        acquire.DeviceKind.Storage, "Tiff"
    )
    assert p.video[0].storage.identifier is not None

    p.video[0].storage.settings.filename = "out.tif"
    assert p.video[0].storage.settings.filename == "out.tif"


def test_setup(runtime: Runtime):
    p = acquire.setup(runtime, "simulated.*empty", "Trash")
    assert p.video[0].camera.identifier is not None
    assert p.video[0].storage.identifier is not None
    assert p.video[0].storage.settings.filename == "out.tif"
    assert p.video[0].max_frame_count == 100
    p.video[0].camera.settings.shape = (192, 108)
    p = runtime.set_configuration(p)

    logging.info(f"max_frame_count: {p.video[0].max_frame_count}")

    runtime.start()

    nframes = 0
    t0 = time.time()

    def took_too_long():
        # Time limit the test
        return time.time() - t0 > 20.0

    while nframes < p.video[0].max_frame_count and not took_too_long():
        clock = time.time()
        if a := runtime.get_available_data(0):
            packet = a.get_frame_count()
            for f in a.frames():
                logging.info(
                    f"{f.data().shape} {f.data()[0][0][0][0]} {f.metadata()}"
                )
                del f  # <-- fails to get the last frames if this is held?
            del a  # <-- fails to get the last frames if this is held?
            nframes += packet
            logging.info(
                f"frame count: {nframes} - frames in packet: {packet}"
            )

        elapsed = time.time() - clock
        sleep(max(0, 0.1 - elapsed))
    logging.info("stopping")

    runtime.stop()
    if took_too_long():
        raise RuntimeError("Took too long")


def test_selection_is_consistent(runtime: Runtime):
    hcam1 = runtime.device_manager().select(DeviceKind.Camera)
    assert hcam1 is not None
    hcam2 = runtime.device_manager().select(DeviceKind.Camera, hcam1.name)
    assert hcam1 == hcam2


def test_change_filename(runtime: Runtime):
    dm = runtime.device_manager()
    p = runtime.get_configuration()
    p.video[0].camera.identifier = dm.select(DeviceKind.Camera, "simulated.*")
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Tiff")
    p.video[0].max_frame_count = 1

    names = [
        "out1.tif",
        "quite a bit longer.tif",
        "s.tif",
        "another long one ok it is really long this time.tif",
    ]
    for name in names:
        p.video[0].storage.settings.filename = name
        p = runtime.set_configuration(p)
        assert p.video[0].storage.settings.filename == name

        nframes = 0
        runtime.start()
        while nframes < p.video[0].max_frame_count:
            if packet := runtime.get_available_data(0):
                nframes += packet.get_frame_count()
                packet = None
        logging.info("Stopping")
        runtime.stop()


def test_write_external_metadata_to_tiff(
    runtime: Runtime, request: pytest.FixtureRequest
):
    dm = runtime.device_manager()
    p = runtime.get_configuration()
    p.video[0].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*sin"
    )
    p.video[0].camera.settings.shape = (33, 47)
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Tiff")
    p.video[0].max_frame_count = 3
    p.video[0].storage.settings.filename = f"{request.node.name}.tif"
    metadata = {"hello": "world"}
    p.video[0].storage.settings.external_metadata_json = json.dumps(metadata)
    runtime.set_configuration(p)

    nframes = 0
    runtime.start()
    while nframes < p.video[0].max_frame_count:
        if packet := runtime.get_available_data(0):
            nframes += packet.get_frame_count()
            packet = None
    runtime.stop()

    # Check that the written tif has the expected structure
    with tifffile.TiffFile(p.video[0].storage.settings.filename) as f:

        def meta(iframe: int) -> Dict[Any, Any]:
            return json.loads(f.pages[iframe].tags["ImageDescription"].value)

        # first frame should have metadata
        assert meta(0)["metadata"] == metadata
        assert meta(0)["frame_id"] == 0

        # remaining frames should not, but should have e.g. frame id
        for i in range(1, p.video[0].max_frame_count):
            assert "metadata" not in meta(i).keys()
            assert meta(i)["frame_id"] == i


@pytest.mark.skipif(sys.platform == "darwin", reason="illegal instruction")
def test_write_external_metadata_to_zarr(
    runtime: Runtime, request: pytest.FixtureRequest
):
    dm = runtime.device_manager()
    p = runtime.get_configuration()
    p.video[0].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*sin.*"
    )
    p.video[0].camera.settings.shape = (33, 47)
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Zarr")
    p.video[0].max_frame_count = 4
    p.video[0].storage.settings.filename = f"{request.node.name}.zarr"
    metadata = {"hello": "world"}
    p.video[0].storage.settings.external_metadata_json = json.dumps(metadata)
    p.video[0].storage.settings.pixel_scale_um = (0.5, 4)

    p = runtime.set_configuration(p)

    nframes = 0
    runtime.start()
    while nframes < p.video[0].max_frame_count:
        if packet := runtime.get_available_data(0):
            nframes += packet.get_frame_count()
            packet = None
    runtime.stop()

    assert p.video[0].storage.settings.filename
    store = parse_url(p.video[0].storage.settings.filename)
    assert store
    reader = Reader(store)
    nodes = list(reader())

    # ome-ngff supports multiple images, in separate directories but we only
    # wrote one.
    multi_scale_image_node = nodes[0]

    # ome-ngff always stores multi-scale images, but we only have a single
    # scale/level.
    image_data = multi_scale_image_node.data[0]
    assert image_data.shape == (
        p.video[0].max_frame_count,
        1,
        p.video[0].camera.settings.shape[1],
        p.video[0].camera.settings.shape[0],
    )

    multi_scale_image_metadata = multi_scale_image_node.metadata

    axes = multi_scale_image_metadata["axes"]
    axis_names = tuple(a["name"] for a in axes)
    assert axis_names == ("t", "c", "y", "x")

    axis_types = tuple(a["type"] for a in axes)
    assert axis_types == ("time", "channel", "space", "space")

    axis_units = tuple(a.get("unit") for a in axes)
    assert axis_units == (None, None, "micrometer", "micrometer")

    # We only have one multi-scale level and one transform.
    transform = multi_scale_image_metadata["coordinateTransformations"][0][0]
    pixel_scale_um = tuple(
        transform["scale"][axis_names.index(axis)] for axis in ("x", "y")
    )
    assert pixel_scale_um == p.video[0].storage.settings.pixel_scale_um

    # ome-zarr only reads attributes it recognizes, so use a plain zarr reader
    # to read external metadata instead.
    group = zarr.open(p.video[0].storage.settings.filename)
    assert group["0"].attrs.asdict() == metadata


@pytest.mark.skipif(sys.platform == "darwin", reason="illegal instruction")
@pytest.mark.parametrize(
    ("storage_kind", "compressor_name", "clevel", "shuffle"),
    [
        ("ZarrBlosc1ZstdByteShuffle", "zstd", 1, blosc.SHUFFLE),
        ("ZarrBlosc1Lz4ByteShuffle", "lz4", 1, blosc.SHUFFLE),
    ],
)
def test_write_compressed_zarr(
    runtime: Runtime,
    request: pytest.FixtureRequest,
    storage_kind: str,
    compressor_name: str,
    clevel: int,
    shuffle: int,
):
    filename = f"{request.node.name}.zarr"
    filename = filename.replace("[", "_").replace("]", "_")

    dm = runtime.device_manager()
    p = runtime.get_configuration()
    p.video[0].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*sin.*"
    )
    p.video[0].camera.settings.shape = (64, 48)
    p.video[0].storage.identifier = dm.select(
        DeviceKind.Storage, "ZarrBlosc1ZstdByteShuffle"
    )
    p.video[0].camera.settings.exposure_time_us = 1e4
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, storage_kind)
    p.video[0].max_frame_count = 70
    p.video[0].storage.settings.filename = filename
    metadata = {"foo": "bar"}
    p.video[0].storage.settings.external_metadata_json = json.dumps(metadata)
    runtime.set_configuration(p)

    runtime.start()
    runtime.stop()

    # load from Zarr
    group = zarr.open(p.video[0].storage.settings.filename)
    data = group["0"]

    assert data.compressor.cname == compressor_name
    assert data.compressor.clevel == clevel
    assert data.compressor.shuffle == shuffle

    assert data.shape == (
        p.video[0].max_frame_count,
        1,
        p.video[0].camera.settings.shape[1],
        p.video[0].camera.settings.shape[0],
    )
    assert data.attrs.asdict() == metadata

    # load from Dask
    data = da.from_zarr(p.video[0].storage.settings.filename, component="0")
    assert data.shape == (
        p.video[0].max_frame_count,
        1,
        p.video[0].camera.settings.shape[1],
        p.video[0].camera.settings.shape[0],
    )


@pytest.mark.skipif(sys.platform == "darwin", reason="illegal instruction")
@pytest.mark.parametrize(
    ("frame_x", "frame_y", "frames_per_chunk", "compressed"),
    [(64, 48, 25, False), (96, 72, 66, False), (1920, 1080, 32, True)],
)
def test_write_raw_zarr_with_variable_chunking(
    runtime: Runtime,
    request: pytest.FixtureRequest,
    frame_x: int,
    frame_y: int,
    frames_per_chunk: int,
    compressed: bool,
):
    max_frames = 100

    dm = runtime.device_manager()

    p = runtime.get_configuration()
    p.video[0].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*empty.*"
    )
    p.video[0].camera.settings.shape = (frame_x, frame_y)
    p.video[0].camera.settings.exposure_time_us = 1e4
    p.video[0].storage.identifier = dm.select(
        DeviceKind.Storage,
        "ZarrBlosc1ZstdByteShuffle" if compressed else "Zarr",
    )
    p.video[0].storage.settings.filename = f"{request.node.name}.zarr"
    p.video[0].max_frame_count = max_frames
    p.video[0].storage.settings.bytes_per_chunk = (
        frame_x * frame_y * frames_per_chunk
    )
    runtime.set_configuration(p)

    n_chunks_expected = int(math.ceil(max_frames / frames_per_chunk))

    runtime.start()
    runtime.stop()

    group = zarr.open(p.video[0].storage.settings.filename)
    data = group["0"]

    assert data.shape == (
        p.video[0].max_frame_count,
        1,
        p.video[0].camera.settings.shape[1],
        p.video[0].camera.settings.shape[0],
    )
    assert data.nchunks == n_chunks_expected


@pytest.mark.skip(
    reason="Runs into memory limitations on github ci."
    + " See https://github.com/acquire-project/cpx/issues/147"
)
def test_two_video_streams(runtime: Runtime):
    dm = runtime.device_manager()
    p = runtime.get_configuration()

    p.video[0].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*random.*"
    )
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    p.video[0].camera.settings.binning = 1
    p.video[0].camera.settings.shape = (64, 64)
    p.video[0].camera.settings.pixel_type = acquire.SampleType.U8
    p.video[0].max_frame_count = 90
    p.video[0].frame_average_count = 0  # disables

    p.video[1].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*empty.*"
    )
    p.video[1].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    p.video[1].camera.settings.binning = 1
    p.video[1].camera.settings.shape = (64, 64)
    p.video[1].camera.settings.pixel_type = acquire.SampleType.U8
    p.video[1].max_frame_count = 71
    p.video[1].frame_average_count = 0  # disables

    p = runtime.set_configuration(p)

    nframes = [0, 0]

    def is_not_done() -> bool:
        return runtime.get_state() == DeviceState.Running and (
            (nframes[0] < p.video[0].max_frame_count)
            or (nframes[1] < p.video[1].max_frame_count)
        )

    runtime.start()

    stream_id = 0
    while is_not_done():
        if nframes[stream_id] < p.video[stream_id].max_frame_count:
            if packet := runtime.get_available_data(stream_id):
                n = packet.get_frame_count()
                for i, frame in enumerate(packet.frames()):
                    expected_frame_id = nframes[stream_id] + i
                    assert frame.metadata().frame_id == expected_frame_id, (
                        "frame id's didn't match "
                        + f"({frame.metadata().frame_id}!={expected_frame_id})"
                        + f" [stream {stream_id} nframes {nframes}]"
                    )
                    del frame
                del packet
                nframes[stream_id] += n
                logging.debug(f"NFRAMES {nframes}")

        stream_id = (stream_id + 1) % 2
    logging.info("Stopping")
    runtime.stop()
    assert nframes[0] == p.video[0].max_frame_count
    assert nframes[1] == p.video[1].max_frame_count


def test_abort(runtime: Runtime):
    dm = runtime.device_manager()
    p = runtime.get_configuration()
    p.video[0].camera.identifier = dm.select(
        DeviceKind.Camera, "simulated.*sin"
    )
    p.video[0].camera.settings.shape = (24, 93)
    p.video[0].storage.identifier = dm.select(DeviceKind.Storage, "Trash")
    p.video[0].max_frame_count = 2**30
    runtime.set_configuration(p)

    nframes = 0
    runtime.start()
    sleep(0.05)
    logging.info("Aborting")
    runtime.abort()

    while packet := runtime.get_available_data(0):
        nframes += packet.get_frame_count()

    del packet

    logging.debug(
        f"Frames expected: {p.video[0].max_frame_count}, actual: {nframes}"
    )
    assert nframes < p.video[0].max_frame_count


# FIXME: (nclack) awkwardness around references  (available frames, f)

# NOTES:
#
# With pytest, use `--log-cli-level=0` to see the lowest level logs.



from typing import List, Optional, Union

import napari  # type: ignore

from .acquire import *

FORMAT: str

def setup(
    runtime: Runtime,
    camera: Union[str, List[str]] = ...,
    storage: Union[str, List[str]] = ...,
    output_filename: Optional[str] = ...,
) -> Properties: ...
def setup_one_streams(runtime: Runtime, frame_count: int) -> Properties: ...
def setup_two_streams(runtime: Runtime, frame_count: int) -> Properties: ...

g_runtime: Optional[Runtime]

def gui(viewer: "napari.Viewer", frame_count: int = ...): ...  # type: ignore
